import java.util.*;
import java.util.stream.*;

/**
 * benchmark created by Aleksey Shipilev in 2012 when the Stream API
 * development was under way; it compares productivity of external (bench_1)
 * and internal (bench_2) computations; the external one has a penalty
 * associated with remembering the state of iterator.
 */

public class LazyEagerBenchmark {
	private static final int COUNT = Integer.getInteger("count", 10_000_000);
	private static List<Integer> list;
	
	public static void setup() {
		list = new ArrayList<>(COUNT);
		for (int c = 1; c <= COUNT; c++) {
			list.add(c);
		}
		Collections.shuffle(list);
	}
	
	public static int bench_1_Iterator() {
		int sum = 0;
		// Iterator<Integer> iter = list.stream() // v1
		Stream<Integer> filter = list.stream() // v2
			.filter(i -> (i & 0xF) == 0)
			// .filter(i -> (i & 0xFF) == 0) // v1
			.filter(i -> (i & 0xFF) == 0);  // v2 (stop mid-way, reuse stream)
		Iterator<Integer> iter = filter     // v2
			.filter(i -> (i & 0xFFF) == 0)
			.filter(i -> (i & 0xFFFF) == 0)
			.iterator(); // v1 and v2 show the same performance
		
		while (iter.hasNext())
			sum += iter.next();
		
		return sum;
	}

	public static int bench_2_ForEach() {
		Counter counter = new Counter();
		list.stream()
			.filter(i -> (i & 0xF) == 0)    // multiple to 16
			.filter(i -> (i & 0xFF) == 0)   // multiple to 256
			.filter(i -> (i & 0xFFF) == 0)  // multiple to 4096
			.filter(i -> (i & 0xFFFF) == 0) // multiple to 65536
			.forEach(counter::add);
		
		return counter.sum;
	}

	private static class Counter {
		public int sum;
		void add(int i) {
			sum += i;
		}
	}
	
	public static void main(String[] args) {
		setup();
		long start = System.currentTimeMillis();
		long res = bench_1_Iterator();
		long finish = System.currentTimeMillis();
		System.out.printf("%d found with iterator, productivity %.5f op/msec%n",
			res, COUNT / (1.0 * (finish - start)));
		setup();
		start = System.currentTimeMillis();
		res = bench_2_ForEach();
		finish = System.currentTimeMillis();
		System.out.printf("%d found with for-each, productivity %.5f op/msec%n",
			res, COUNT / (1.0 * (finish - start)));
	}
}